/*****************************************************************************/
/* fizip.c								     */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* Command line user interface to compress/decompress FITS data.	     */
/*****************************************************************************/
#define	FI_ZIP_VERSION		"0.2"
/*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdarg.h>
#include <libgen.h>

#include <fits/fits.h>

#include "longhelp.h"
#include "fi.h"

#include "io/iof.h"
#include "io/scanarg.h"
#include "io/tokenize.h"

#include "zip.h"

#ifdef  HAVE_NO_CC_EXTENSION 
#define __extension__ 
#endif 

/*****************************************************************************/

int	is_verbose,is_comment;
char	*progbasename;

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

int fprint_error(char *expr,...)
{
 va_list	ap;
 fprintf(stderr,"%s: error: ",progbasename);
 va_start(ap,expr);
 vfprintf(stderr,expr,ap);
 va_end(ap);
 fprintf(stderr,"\n");
 return(0);
}

int fprint_warning(char *expr,...)
{
 va_list	ap;
 fprintf(stderr,"%s: warning: ",progbasename);
 va_start(ap,expr);
 vfprintf(stderr,expr,ap);
 va_end(ap);
 fprintf(stderr,"\n");
 return(0);
}

/*****************************************************************************/

int fprint_fizip_usage(FILE *fw)
{
 fprintf(fw,
"Usage:\tfizip|fiunzip [--help] [<in>[<F>] [-o|--output <out>]\n"
"\t[-z|--compress] [-d|--decompress]\n"
"\t[--quantizebits <N> [--round] | --roundbits <N>]\n"
"Compression methods:\n"
#ifdef	ENABLE_BZ2
"\t--bzip2 [--bzip2-tune blocksize=<1...9>,workfactor=<0...250>]\n"
#endif
#ifdef	ENABLE_GZ
"\t--gzip  [--gzip-tune compression=<-1,1...9>]\n"
#endif
"\t--none\n"
 );

 fprintf(fw,
"For a more detailed description of the meanings of the --*-tune arguments,\n" 
"see the manual pages of `bzip2` and `gzip` programs. The compression method\n"
"is also stored in the FITS headers and cannot be forced to be changed\n"
"during decompression. \n");

 return(0);
}

longhelp_entry fizip_long_help[]=
{
 { "General options:", NULL },
 { "-h, --help",
        "Give general summary about the command line options." },
 { "--long-help",
        "Give a detailed list of command line options." },
 { "--version",
        "Give some version information about the program." },
 { "<input>",
	"Name of the input FITS image to be compressed or decompress. "
	"By default, fi[un]zip writes the compressed/decompressed image "
	"data to the standard output. " },
 { "-o, --output <output>",
	"Name of the output FITS image. Can be the same as the input image "
	"file name, in this case  the original file will be overwritten with "
	"its compressed/decompressed form." },
 { "-z, --compress",
	"Compress the input image. This is the default when the program is "
	"called as `fizip`." },
 { "-d, --decompress",
	"Decompress the input image. This is the default when the program is "
	"called as `fiunzip`. If the program is invoked not as `fizip` or "
	"`fiunzip` and neither -z nor -d is specified, the behaviour of "
	"the program is ambiguous and therefore it will abort." },
 { "--quantizebits <N>",
	"Do a quantization (rounding down) before compression "
	"on the upper <N> bits of the ``decimals''. "
	"If <N> is zero, all of the the pixel values will be rounded down "
	"to the largest integer. If <N> is negative, the pixel values will be "
	"rounded down to the largest multiple of the -<N>th power of 2. " },
 { "--roundbits <N>",
	"Do a rounding before compression "
	"on the upper <N> bits of the ``decimals''. "
	"If <N> is zero, all of the the pixel values will be rounded "
	"to the closest integer. If <N> is negative, the pixel values will be "
	"rounded to the closest multiple of the -<N>th power of 2. " },
 { "--round",
	"Equivalent to ``--roundbits <N>'' if ``--quantizebits <N>'' is used. " },

 { "Compression algorithms:" },
 { "--none",
	"Do not compress, just store the raw data in the same kind of format. " },
 { "--gzip",
	"Use the GNU zip library for compression." },
 { "--gzip-tune compression=<level>",
	"Tune the GNU zip compression algorithm, see `man gzip` for a detailed "
	"explanation of the meaning of the available options. " },
 { "--bzip2",
	"Use the block-sorting file compressor (bzip2) library "
	"for compression." },
 { "--bzip2-tune blocksize=<1...9>,workfactor=<0...250>",
	"Tune the bzip2 compression algorithm, see `man bzip2` for a detailed "
	"explanation of the meaning of the available options. " },

 {	"Note that the above list reflects the implemented list of compression "
	"algorithms, the actually available algorithms depend on the build "
	"environment, i.e. the libz-dev or lipbzip2-dev development libraries "
	"must be intalled on the host at the compilation time of `fizip` for "
	"a support ot the GNU zip and bzip2 algorithms, respectively. "
	"The dummy compression ``--none'' is always available.", NULL }, {"",NULL},

 { NULL, NULL }

};

int fprint_fizip_long_help(FILE *fw)
{
 fprintf(fw,
"Usage:\tfi[un]zip [<compression options> | -d] <input> [-o <output>]\n");

 fprintf(fw,
"The puropse of the `fizip` and `fiunzip` programs is to compress and \n"
"decompress primary FITS data, by keeping the changes in the primary FITS\n"
"header to be minimal. The compressed data is stored in a one-dimensional\n"
"8 bit (BITPIX=8, NAXIS=1) array, therefore these keywords does not\n");
 fprintf(fw,
"reflect the original image dimension or data type. All of the other keywords\n"
"are untouched. Some auxiliary information on the compression is stored in\n"
"the keywords starting with ``FIZIP'', the contents of these keywords\n"
"depend on the involved compression method. `fizip` rejects compressing\n"
"FITS file where such keywords exist in the primary header.\n\n");

 longhelp_fprint(fw,fizip_long_help,0,-1);

 fprintf(fw,"\n");
 fprintf(fw,"Report bugs to <%s>\n",FI_MAINT_EMAIL);

 return(0);
}

/*****************************************************************************/

errormessage fizip_errmsgs[]=
 {	{ FIZIP_ERR_INTERNAL,
		"internal error (please send a bug-report)"	},
	{ FIZIP_ERR_ALREADY_COMPRESSED,
		"FITS data seem to be compressed"		},
	{ FIZIP_ERR_UNIMPLEMENTED_METHOD,
		"unimplemented (de)compression method"		},
	{ FIZIP_ERR_UNSUPPORTED_EXTENSION,
		"unsupported FITS extension in data"		},
	{ FIZIP_ERR_UNEXPECTED_EOF,
		"unexpected end-of-file occured"		},
	{ FIZIP_ERR_COMPRESS, 
		"raw compression failed"			},
	{ FIZIP_ERR_DECOMPRESS,
		"raw decompression failed"			},
	{ FIZIP_ERR_CONFLICTING_HEADERS,
		"FITS headers have conflicting values"		},
	{ FIZIP_ERR_NOT_COMPRESSED,
		"FITS data seem not to be compressed"		},
	{ 0,NULL }
 };

char *src_errormessage(errormessage *eml,int code)
{
 while ( eml->message != NULL )
  {	if ( eml->code == code )	return(eml->message);
	eml++;
  };
 return(NULL);
}

/*****************************************************************************/

int filecmp(char *f1,char *f2)
{
 char	*f1d1,*f1d2,*f2d1,*f2d2,
	*f1base,*f1dir,*f2base,*f2dir;
 int	i;
 if ( f1==NULL && f2==NULL )		return(0);
 else if ( f1==NULL || f2==NULL )	return(1);
 f1d1=strdup(f1);
 for ( i=0 ; f1d1[i] ; i++ )
  {	if ( f1d1[i]=='/' && f1d1[i+1]=='/' )
		memmove(f1d1+i,f1d1+i+1,strlen(f1d1+i));
  }
 f1d2=strdup(f1d1);
 f2d1=strdup(f2);
 for ( i=0 ; f2d1[i] ; i++ )
  {	if ( f2d1[i]=='/' && f2d1[i+1]=='/' )
		memmove(f2d1+i,f2d1+i+1,strlen(f2d1+i));
  }
 f2d2=strdup(f2d1);
 f1base=basename(f1d1);
 f1dir =dirname(f1d2);
 f2base=basename(f2d1);
 f2dir =dirname(f2d2);
 if ( strcmp(f1base,f2base)==0 && strcmp(f1dir,f2dir)==0 )
	i=0;
 else
	i=1;
 free(f2d2);free(f2d1);free(f1d2);free(f1d1);
 return(i);
}

/*****************************************************************************/

typedef struct
 {	void	*buff;	
	int	size;
 } memblock;

int cb_write_to_memory(void *param,void *buff,int len)
{
 memblock	*mb;
 mb=(memblock *)param;
 mb->buff=realloc(mb->buff,mb->size+len);
 memcpy(((char *)mb->buff)+mb->size,buff,len);
 mb->size+=len;
 return(len);
}

/*****************************************************************************/

int main(int argc,char *argv[])
{
 fits		*img;

 FILE		*fr,*fw;
 int		i,is_help,compress_dir,ret,via_memory;
 int		compress_method;
 void		*compress_param;
 char		*file_in,*file_out,*basename,*parambz2,*paramgz;
 int		(*cb_read)(void *,void *,int);
 int		(*cb_write)(void *,void *,int);
 void		*writeparam,*readparam;

#ifdef	ENABLE_BZ2
 compressbz2	cbz2;
#endif
#ifdef	ENABLE_GZ
 compressgz	cgz;
#endif

 compresstune	ct;
 memblock	outmem;

 progbasename=strrchr(argv[0],'/');
 if ( progbasename != NULL )	progbasename++;
 else				progbasename=argv[0];

 is_help=is_verbose=is_comment=0;

 file_in=file_out=NULL;

 compress_dir=-1;
#if defined	ENABLE_BZ2
 compress_method=COMPRESS_BZ2; 
#elif defined	ENABLE_GZ
 compress_method=COMPRESS_GZ;
#else
 compress_method=COMPRESS_NONE;
#endif

 parambz2=NULL;
 paramgz =NULL;

 ct.nquantizebit=-1;
 ct.is_round=0;

 i=scanarg(argc,argv,SCANARG_ALLOW_FLAGS,
 	"--version:%NS-1f%q",&is_help,
        "--version-short|--short-version:%NS-2f%q",&is_help,
	"-h|--help|--short-help|--help-short:%f%q",&is_help,
        "--long-help|--help-long:%SN2f%q",&is_help,
	"-i|--input:%s",&file_in,
	"-o|--output:%s",&file_out,
	"-z|--compress:%SN0f",&compress_dir,
	"-d|--decompress:%SN1f",&compress_dir,

	"--none:"	 SNf(COMPRESS_NONE),&compress_method,

#ifdef	ENABLE_BZ2
	"--bz2|--bzip2:" SNf(COMPRESS_BZ2 ),&compress_method,
	"--bz2-tune|--bzip2-tune:%s",&parambz2,
#endif
#ifdef	ENABLE_GZ
	"--gz|--gzip:"   SNf(COMPRESS_GZ  ),&compress_method,
	"--gz-tune|--gzip-tune:%s",  &paramgz,
#endif
	"--quantizebits:%ZPd",&ct.nquantizebit,
	"--roundbits:%ZPd%f",&ct.nquantizebit,&ct.is_round,
	"--round:%f",&ct.is_round,
	"--verbose:%i",&is_verbose,"(V):%i",&is_verbose,
	"-:%w",&file_in,
	"-*|+*:%e",
	"*:%w",&file_in,
	NULL);

 if ( i )		
  {	fprint_error("invalid command line argument near '%s'",argv[i]);
	return(1);
  }
 else if ( is_help<0 )
  {	fprint_generic_version(stdout,argv[0],"fizip",FI_ZIP_VERSION,is_help);
	return(0);
  }
 else if ( is_help>1 )
  {	fprint_fizip_long_help(stdout);
	return(0);
  }
 else if ( is_help )
  {	fprint_fizip_usage(stdout);
	return(0);
  }

 basename=strrchr(argv[0],'/');
 if ( basename==NULL )	basename=argv[0];
 else			basename=basename+1;
 if ( compress_dir<0 )
  {	if ( strcasecmp(basename,"fizip")==0 )
		compress_dir=0;
	else if ( strcasecmp(basename,"fiunzip")==0 )
		compress_dir=1;
  }

 if ( compress_dir<0 )
  {	fprint_error("unable to detect compression direction\n");
 	return(1);
  }

 compress_param=NULL;

#ifdef	ENABLE_BZ2
 if ( parambz2 != NULL )
  {	if ( compress_method != COMPRESS_BZ2 )
	 {	fprint_error("invalid combination of command line arguments");
		return(1);
	 }
	i=scanpar(parambz2,SCANPAR_DEFAULT,
		"blocksize:%d",&cbz2.blocksize,
		"workfactor:%d",&cbz2.workfactor,
		NULL);
	if ( i )	
	 {	fprint_error("invalid bzip2 fine-tune argument in '%s'",parambz2);
		return(1);
	 }
	compress_param=&cbz2;
  }
#endif

#ifdef	ENABLE_GZ
 if ( paramgz != NULL )
  {	if ( compress_method != COMPRESS_GZ )
	 {	fprint_error("invalid combination of command line arguments");
		return(1);
	 }
	i=scanpar(paramgz,SCANPAR_DEFAULT,
		"compression:%d",&cgz.compression,
		NULL);
	if ( i )
	 {	fprint_error("invalid gzip fine-tune argument in '%s'",parambz2);
		return(1);
	 }
	compress_param=&cgz;
  }
#endif

 if ( file_in==NULL || file_out==NULL )
	via_memory=0;
 else if ( strcmp(file_in,"-")==0 || strcmp(file_out,"-")==0 )
	via_memory=0;
 else if ( filecmp(file_in,file_out)==0 )
	via_memory=1;
 else
	via_memory=0;

 if ( file_in==NULL )	fr=stdin;
 else			fr=fopenread(file_in);
 if ( fr==NULL )	
  {	fprint_error("unable to open input file '%s'",file_in);
	return(1);
  }

 img=fits_create();
 fits_headerset_reset(&img->header);

 cb_read =fits_cb_read;		/* the built-in callbacks can be used so */
 readparam=(void *)fr;
 
 if ( via_memory )
  {	outmem.buff=NULL;
	outmem.size=0;
	cb_write=cb_write_to_memory;
	writeparam=(void *)(&outmem);
	fw=NULL;
  }
 else
  {	if ( file_out==NULL )	fw=stdout;
	else			fw=fopenwrite(file_out);
	if ( fw==NULL )		
	 {	fprint_error("unable to create output file '%s'",file_out);
		return(1);
	 }
	cb_write=fits_cb_write;
	writeparam=(void *)fw;
  }

 if ( ! compress_dir )
  {	ret=fizip_compress(cb_read,readparam,cb_write,writeparam,
		compress_method,compress_param,&ct);
	if ( ret )
	 {	fprint_error("during compression: %s",src_errormessage(fizip_errmsgs,ret));
		return(1);
	 }
  }
 else
  {	ret=fizip_decompress(cb_read,readparam,cb_write,writeparam);
	if ( ret )
	 {	fprint_error("during decompression: %s",src_errormessage(fizip_errmsgs,ret));
		return(1);
	 }
  }

 if ( via_memory )
  {	if ( file_out==NULL )	fw=stdout;
	else			fw=fopenwrite(file_out);
	if ( fw==NULL )		
	 {	fprint_error("unable to create output file '%s'",file_out);
		return(1);
	 }
	fwrite(outmem.buff,1,outmem.size,fw);
	fclosewrite(fw);
  }
 else
  {	if ( fw != NULL )	fclosewrite(fw);	}

 fcloseread(fr);

 return(0);
}
