/*****************************************************************************/
/* zip.c								     */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* An API for transparent and standard-preserving FITS compression.	     */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* (c) 2006; Pal, A. (apal@szofi.elte.hu)				     */
/*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdarg.h>

#include <fits/fits.h>

#include "zip.h"

#ifdef	ENABLE_BZ2
#include <bzlib.h>
#endif

#ifdef	ENABLE_GZ
#include <zlib.h>
#endif

/*****************************************************************************/

#define	HDR_FIZIP		0
#define	HDR_FIZIPMTD		1
#define	HDR_FIZIPVRS		2
#define	HDR_FIZIPHDR		3

char *fiziphdrs[]=
 {	"FIZIP",
	"FIZIPMTD",
	"FIZIPVRS",
	"FIZIPHDR",
	NULL
 };

/*****************************************************************************/

compressinfo complist[]=
 {	
	{ COMPRESS_NONE,"NONE",	fizip_none_version,
				fizip_none_compress,
				fizip_none_decompress	},

#ifdef	ENABLE_BZ2
	{ COMPRESS_BZ2, "BZ2",	fizip_bz2_version, 
				fizip_bz2_compress,
				fizip_bz2_decompress	},
#endif
#ifdef	ENABLE_GZ
	{ COMPRESS_GZ,	"GZ",	fizip_gz_version,
				fizip_gz_compress,
				fizip_gz_decompress	},
#endif

	{ -1, NULL, NULL, NULL, NULL }
 };

/*****************************************************************************/

char *fizip_none_version(void)
{
 return("0.0");
}
int fizip_none_compress(char *in,int insize,char *out,int *outsize,void *param)
{
 if ( in==NULL || out==NULL || outsize==NULL )	return(FIZIP_ERR_INTERNAL);

 if ( *outsize<insize )
 	return(1);	
 else
  {	memcpy(out,in,insize);
	*outsize=insize;
	return(0);
  }
}
int fizip_none_decompress(char *in,int insize,char *out,int *outsize)
{
 if ( in==NULL || out==NULL || outsize==NULL )	return(FIZIP_ERR_INTERNAL);

 if ( *outsize<insize )
 	return(1);	
 else
  {	memcpy(out,in,insize);
	*outsize=insize;
	return(0);
  }
}

/*****************************************************************************/

#ifdef	ENABLE_BZ2

/* libbz2: wrappers to the bzip/bunzip library functions */
char *fizip_bz2_version(void)
{
 return((char *)BZ2_bzlibVersion());
}
int fizip_bz2_compress(char *in,int insize,char *out,int *outsize,void *param)
{
 int		ret;
 int		blocksize,workfactor;
 compressbz2 	*cp;

 if ( in==NULL || out==NULL || outsize==NULL )	return(FIZIP_ERR_INTERNAL);

 if ( param==NULL )
  {	blocksize=1,
	workfactor=0;
  }
 else	
  {	cp=(compressbz2 *)param;
	blocksize=cp->blocksize;
	workfactor=cp->workfactor;
  }
 if ( blocksize<1 )		blocksize=1;
 else if ( blocksize>9 )	blocksize=9;
 if ( workfactor<0 )		workfactor=0;
 else if ( workfactor>250 )	workfactor=250;

 ret=BZ2_bzBuffToBuffCompress(out,(unsigned int *)outsize,in,insize,blocksize,0,workfactor);

 if ( ret==BZ_OK )			return(0);
 else if ( ret==BZ_OUTBUFF_FULL )	return(1);
 else					return(FIZIP_ERR_INTERNAL);
}
int fizip_bz2_decompress(char *in,int insize,char *out,int *outsize)
{
 int	ret;
 if ( in==NULL || out==NULL || outsize==NULL )	return(FIZIP_ERR_INTERNAL);
 ret=BZ2_bzBuffToBuffDecompress(out,(unsigned int *)outsize,in,insize,0,0);
 if ( ret==BZ_OK )			return(0);
 else if ( ret==BZ_OUTBUFF_FULL )	return(1);
 else					return(FIZIP_ERR_INTERNAL);
}
#endif

/*****************************************************************************/

#ifdef	ENABLE_GZ

#define		GZ_MAGIC_1	0x1F
#define		GZ_MAGIC_2	0x8B

int fizip_gz_set_uint(char *buff,uInt val)
{
 int	i;
 uLong	v;

 v=(uLong)val;
 for ( i=0 ; i<4 ; i++ )
  {	*buff = (v & 0xff);
	buff++;
	v >>= 8;
  }
 
 return(0);
}

uInt fizip_gz_get_uint(char *buff)
{
 int	i;
 uLong	v;
 v=0;
 for ( i=0 ; i<4 ; i++ )
  {	v += ((unsigned char)*buff) << (i*8);
	buff++;
  }
 return((uInt)v);
}

/* libbz2: wrappers to the bzip/bunzip library functions */
char *fizip_gz_version(void)
{
 return((char *)zlibVersion());
}

int fizip_gz_compress(char *in,int insize,char *out,int *outsize,void *param)
{
 int		compression,totalout;
 compressgz 	*cp;
 z_stream	stream;
 uInt		crc;

 if ( in==NULL || out==NULL || outsize==NULL )	return(FIZIP_ERR_INTERNAL);

 if ( param==NULL )
  {	compression=Z_DEFAULT_COMPRESSION;
  }
 else	
  {	cp=(compressgz *)param;
	compression=cp->compression;
  }
 if ( compression<0 )		compression=-1;
 else if ( compression<1 )	compression=1;
 else if ( compression>9 )	compression=9;

 if ( *outsize<=10 )	return(1);	/* no space for ".gz" header */
 sprintf(out,"%c%c%c%c%c%c%c%c%c%c",GZ_MAGIC_1,GZ_MAGIC_2,Z_DEFLATED,0,0,0,0,0,0,3);
 out+=10;(*outsize)-=10;
 totalout=10;

 stream.next_in=(Bytef *)in;
 stream.avail_in=(uInt)insize;
 stream.next_out=(Bytef *)out;
 stream.avail_out=(uInt)(*outsize);

 stream.zalloc=NULL;
 stream.zfree =NULL;
 stream.opaque=NULL;

 deflateInit2(&stream,compression,Z_DEFLATED,-MAX_WBITS,8,Z_DEFAULT_STRATEGY);
 deflate(&stream,Z_FINISH);
 deflateEnd(&stream);

 totalout+=stream.total_out;
 (*outsize)-=stream.total_out;
 out+=stream.total_out;

 if ( *outsize<8 )	return(1);	/* no space for ".gz" trailer */
 
 crc=crc32(0L,(const Bytef *)in,insize);
 
 fizip_gz_set_uint(out+0,crc);
 fizip_gz_set_uint(out+4,insize);

 totalout+=8;

 *outsize=totalout;

 return(0);
}
int fizip_gz_decompress(char *in,int insize,char *out,int *outsize)
{
 int		chkinsize;
 z_stream	stream;
 uInt		crc;

 if ( insize<10	)
	return(1);
 if ( (unsigned char)in[0] != GZ_MAGIC_1 || (unsigned char)in[1] != GZ_MAGIC_2 )
	return(1);

 in+=10;insize-=10;

 stream.next_in=(Bytef *)in;
 stream.avail_in=(uInt)insize-8;
 stream.next_out=(Bytef *)out;
 stream.avail_out=(uInt)(*outsize);

 stream.zalloc=NULL;
 stream.zfree =NULL;
 stream.opaque=NULL;

 inflateInit2(&stream,-MAX_WBITS);
 inflate(&stream,Z_FINISH);
 inflateEnd(&stream);

 in+=insize-8;
 
 crc	  =fizip_gz_get_uint(in+0);
 chkinsize=fizip_gz_get_uint(in+4);

 if ( chkinsize != stream.total_out )
	return(2);
 
 *outsize=stream.total_out;
 
 if ( crc != crc32(0L,(Bytef *)out,*outsize) )
	return(3);
 
 return(0);
}
#endif

/*****************************************************************************/

compressinfo * fizip_get_compressinfo(int method)
{
 compressinfo	* ci;

 for ( ci=complist ; ci->type>=0 ; ci++ )
  {	if ( ci->type==method )		break;		}
 if ( ci->type<0 )	return(NULL);
 else			return(ci);
}

compressinfo * fizip_get_compressinfo_byname(char *name)
{
 compressinfo	* ci;

 for ( ci=complist ; ci->type>=0 ; ci++ )
  {	if ( strcmp(ci->name,name)==0 )		break;		}
 if ( ci->type<0 )	return(NULL);
 else			return(ci);
}

/*****************************************************************************/

int fizip_save_image_headers(fitsimage *fi,char *hbuff)
{
 sprintf(hbuff,"NAXIS=%d,NAXIS1=%d,BITPIX=%d",
	fi->dim,fi->naxis[0],fi->bit);
 return(0);
}
int fizip_save_table_headers(fitsttable *ft,char *hbuff)
{ 
 sprintf(hbuff,"NAXIS1=%d,NAXIS2=%d,TFIELDS=%d,TBCOL1=%d,TFORM1=%s",
	ft->rowsize,ft->nrow,ft->ntfield,ft->tfields[0].colindex+1,ft->tfields[0].format);
 return(0);
}
int fizip_save_bintable_headers(fitsbtable *fb,char *hbuff)
{
 sprintf(hbuff,"NAXIS1=%d,NAXIS2=%d,TFIELDS=%d,TFORM1=%c%d",
	fb->rowsize,fb->nrow,fb->nbfield,fb->bfields[0].form,fb->bfields[0].repeat);
 return(0);
}

int fizip_check_number(char *hbuff)
{
 while ( *hbuff && *hbuff != ',' && ! isspace(*hbuff) )
  {	if ( ! ( isdigit(*hbuff) || *hbuff=='-' || *hbuff=='+' ) )
		return(0);
	hbuff++;
  };
 return(1);
}

int fizip_restore_headers(fitsheaderset *header,char *hbuff)
{
 char		hkwd[16],hstr[64];
 int		i,hval;
 fitsheader	*hd;

 while ( *hbuff )
  {	i=0;
	while ( isblank(*hbuff) )	hbuff++;
	while ( *hbuff && *hbuff != '=' && i<8 )
	 {	hkwd[i]=*hbuff,hbuff++,i++;	}
	hkwd[i]=0;
	if ( *hbuff=='=' )	hbuff++;
	while ( isblank(*hbuff) )	hbuff++;
	if ( fizip_check_number(hbuff) )
	 {	sscanf(hbuff,"%d",&hval);
		hd=fits_headerset_get_header(header,hkwd,0);
		hd->vtype=FITS_VINT;
		hd->vint =hval;
	 }
	else	/* here we don't check for spaces with isspace() ... */
	 {	i=0;
		while ( *hbuff && *hbuff != ',' && i<63 )
		 {	hstr[i]=*hbuff,hbuff++,i++;		}
		hstr[i]=0;
		hd=fits_headerset_get_header(header,hkwd,0);
		hd->vtype=FITS_VSTR;
		strcpy(hd->vstr,hstr);
	 }

	while ( *hbuff && *hbuff != ',' )	hbuff++;
	if ( *hbuff==',' )	hbuff++;
  };

 return(0);
}

/** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** **/

int fizip_arch_is_swapped(void)
{
 short  a;
 a=1;
 if ( *(char *)(&a) == 1 )      return(1);
 else                           return(0);
}

int fizip_quantize_image(char *imgrd,int ntot,int bs,int nquantizebit,int is_round)
{
 int	is_swapped;
 float	wf;
 double	wd,pw,ipw;
 char	wn[8];

 if ( imgrd==NULL || ntot<=0 || ! ( bs==4 || bs==8 ) || nquantizebit<0 )
	return(0);
 
 is_swapped=fizip_arch_is_swapped();

 pw=(double)(1<<(nquantizebit));
 ipw=1.0/pw;

 while ( ntot>0 )
  {	if ( is_swapped )
	 {	switch ( bs )
		 {   case 4:
			wn[0]=imgrd[3],wn[1]=imgrd[2],wn[2]=imgrd[1],wn[3]=imgrd[0];
			wf=*(float *)(&wn[0]);
			wd=wf;
			break;
		     case 8:
			wn[0]=imgrd[7],wn[1]=imgrd[6],wn[2]=imgrd[5],wn[3]=imgrd[4];
			wn[4]=imgrd[3],wn[5]=imgrd[2],wn[6]=imgrd[1],wn[7]=imgrd[0];
			wd=*(double *)(&wn[0]);
			break;
		     default:
			wd=0.0;
			break;
		 }
	 }
	else
	 {	switch ( bs )
		 {   case 4:
			wf=*(float *)imgrd;
			wd=wf;
			break;
		     case 8:
			wd=*(double *)imgrd;
			break;
		     default:
			wd=0.0;
			break;
		 }
	 }

	if ( is_round )	wd=ipw*floor(wd*pw+0.5);
	else		wd=ipw*floor(wd*pw);

	if ( is_swapped )
	 {	switch ( bs )
		 {   case 4:
			wf=wd;
			*(float *)(&wn[0])=wf;
			imgrd[3]=wn[0],imgrd[2]=wn[1],imgrd[1]=wn[2],imgrd[0]=wn[3];
			break;
		     case 8:
			*(double *)(&wn[0])=wd;
			imgrd[7]=wn[0],imgrd[6]=wn[1],imgrd[5]=wn[2],imgrd[4]=wn[3];
			imgrd[3]=wn[4],imgrd[2]=wn[5],imgrd[1]=wn[6],imgrd[0]=wn[7];
			break;
		 }
	 }
	else
	 {	switch ( bs )
		 {   case 4:
			wf=wd;
			*(float *)imgrd=wf;
			break;
		     case 8:
			*(double *)imgrd=wd;
			break;
		 }
	 }
	
	imgrd+=bs;
	ntot--;
  };
	
 return(0); 
}

int fizip_compress_image(int (*cb_read )(void *,void *,int),void *rdparam,
			 int (*cb_write)(void *,void *,int),void *wrparam,
			fitsheaderset *header,fitsimage *fi,compressinfo *ci,
			void *param,compresstune *ct)
{
 char		hbuff[80];
 int		bs,ntot,btot,ctot,cmps,wtot,nread;
 fitsheader	*hd;
 char		*imgrd,*imgcc;

 if ( header==NULL || fi==NULL || ci==NULL || cb_read==NULL || cb_write==NULL )
	return(FIZIP_ERR_INTERNAL);

 fizip_save_image_headers(fi,hbuff);
 fits_headerset_set_string(header,fiziphdrs[HDR_FIZIPHDR],FITS_SH_FIRST,hbuff,NULL);
 bs=fi->bit;
 if ( bs<0 )	bs=-bs;
 bs=bs/8;
 ntot=fits_image_total_pixels(fi->dim,fi->naxis);
 btot=ntot*bs;
 ctot=fits_tapeblock_size(btot);

 imgrd=(char *)malloc(ctot);
 nread=cb_read(rdparam,imgrd,ctot);
 if ( nread < ctot )
	return(FIZIP_ERR_UNEXPECTED_EOF);
 if ( ct != NULL && ct->nquantizebit>=0 && fi->bit<0 )
  {	fizip_quantize_image(imgrd,ntot,bs,ct->nquantizebit,ct->is_round);	}

 cmps=ctot;
 imgcc=(char *)malloc(cmps);

 if ( ci->compress(imgrd,btot,imgcc,&cmps,param) )
	return(FIZIP_ERR_COMPRESS);
 
 hd=fits_headerset_get_header(header,"NAXIS",0);
 hd->vint=1;
 hd=fits_headerset_get_header(header,"NAXIS1",0);
 hd->vint=cmps;
 hd=fits_headerset_get_header(header,"BITPIX",0);
 hd->vint=8;

 fits_headerset_write_cb(cb_write,wrparam,header);
 
 wtot=fits_tapeblock_size(cmps);
 imgcc=(char *)realloc(imgcc,wtot);
 if ( wtot>cmps )	memset(imgcc+cmps,0,wtot-cmps);

 cb_write(wrparam,imgcc,wtot);

 return(0);
}

int fizip_decompress_image(int (*cb_read )(void *,void *,int),void *rdparam,
			   int (*cb_write)(void *,void *,int),void *wrparam,
			   fitsheaderset *header,char *hbuff,compressinfo *ci)
{
 fitsheader	*hd;
 char		*imgcc,*imgrd;
 int		cmtot,crtot,bs,ntot,ctot,atot,rtot,w;
 fitsimage	fi;

 hd=fits_headerset_get_uniq_header(header,"NAXIS");
 if ( hd==NULL || hd->vtype != FITS_VINT || hd->vint != 1 )
 	return(FIZIP_ERR_CONFLICTING_HEADERS);
		/* discrepancy: NAXIS is not present or it is not 1  */
 hd=fits_headerset_get_uniq_header(header,"NAXIS1");
 if ( hd==NULL || hd->vtype != FITS_VINT )
	return(FIZIP_ERR_CONFLICTING_HEADERS);
		/* discrepancy: NAXIS1 is not present		     */

 cmtot=hd->vint;
 crtot=fits_tapeblock_size(cmtot);

 fizip_restore_headers(header,hbuff);

 if ( fits_image_get_params(header,&fi) )
	return(FIZIP_ERR_CONFLICTING_HEADERS);
 bs=fi.bit;
 if ( bs<0 )	bs=-bs;
 bs=bs/8;
 ntot=fits_image_total_pixels(fi.dim,fi.naxis);
 ctot=ntot*bs;
 atot=ntot*bs+1;
 rtot=fits_tapeblock_size(ctot);
 
 imgrd=(char *)malloc(rtot);
 imgcc=(char *)malloc(crtot);

 cb_read(rdparam,imgcc,crtot);
 w=ci->decompress(imgcc,cmtot,imgrd,&atot);

 free(imgcc);
 /* discrepancy: the uncompressed size of the		*/
 /* buffer is not what is expected from the NAXIS*	*/
 /* and BITPIX headers of the image!			*/
 if ( w || atot != ctot )
  {	free(imgrd);
	return(FIZIP_ERR_DECOMPRESS);
  }

 if ( ctot<rtot )	memset(imgrd+ctot,0,rtot-ctot);

 fits_headerset_write_cb(cb_write,wrparam,header);
 cb_write(wrparam,imgrd,rtot);

 free(imgrd);
	
 return(0);
}

/** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** **/

int fizip_compress_table(int (*cb_read )(void *,void *,int),void *rdparam,
			 int (*cb_write)(void *,void *,int),void *wrparam,
			fitsheaderset *header,fitsttable *ft,compressinfo *ci,
			void *param,compresstune *ct)
{
 char		hbuff[80];
 int		btot,ctot,cmps,wtot,nread;
 fitsheader	*hd;
 char		*imgrd,*imgcc;

 if ( header==NULL || ft==NULL || ci==NULL || cb_read==NULL || cb_write==NULL )
	return(FIZIP_ERR_INTERNAL);

 fizip_save_table_headers(ft,hbuff);
 fits_headerset_set_string(header,fiziphdrs[HDR_FIZIPHDR],FITS_SH_FIRST,hbuff,NULL);
 btot=ft->nrow*ft->rowsize;

 ctot=fits_tapeblock_size(btot);

 imgrd=(char *)malloc(ctot);
 nread=cb_read(rdparam,imgrd,ctot);
 if ( nread < ctot )	
	return(FIZIP_ERR_UNEXPECTED_EOF);
 cmps=ctot;
 imgcc=(char *)malloc(cmps);

 if ( ci->compress(imgrd,btot,imgcc,&cmps,param) )
	return(FIZIP_ERR_COMPRESS);
 
 hd=fits_headerset_get_header(header,"NAXIS1",0);
 hd->vint=1;
 hd=fits_headerset_get_header(header,"NAXIS2",0);
 hd->vint=cmps;
 hd=fits_headerset_get_header(header,"TFIELDS",0);
 hd->vint=1;
 hd=fits_headerset_get_header(header,"TBCOL1",0);
 hd->vint=1;
 hd=fits_headerset_get_header(header,"TFORM1",0);
 strcpy(hd->vstr,"A1");

 fits_headerset_write_cb(cb_write,wrparam,header);
 
 wtot=fits_tapeblock_size(cmps);
 imgcc=(char *)realloc(imgcc,wtot);
 if ( wtot>cmps )	memset(imgcc+cmps,0,wtot-cmps);

 cb_write(wrparam,imgcc,wtot);

 return(0);
}

int fizip_decompress_table(int (*cb_read )(void *,void *,int),void *rdparam,
			   int (*cb_write)(void *,void *,int),void *wrparam,
			   fitsheaderset *header,char *hbuff,compressinfo *ci)
{
 fitsheader	*hd;
 char		*imgcc,*imgrd;
 int		cmtot,crtot,ctot,atot,rtot,w;
 fitsttable	ft;

 hd=fits_headerset_get_uniq_header(header,"NAXIS");
 if ( hd==NULL || hd->vtype != FITS_VINT || hd->vint != 2 )
 	return(FIZIP_ERR_CONFLICTING_HEADERS);
		/* discrepancy: NAXIS is not present or it is not 2  */
 hd=fits_headerset_get_uniq_header(header,"NAXIS1");
 if ( hd==NULL || hd->vtype != FITS_VINT || hd->vint != 1 )
	return(FIZIP_ERR_CONFLICTING_HEADERS);
		/* discrepancy: NAXIS1 is not present or it is not 1 */
 hd=fits_headerset_get_uniq_header(header,"NAXIS2");
 if ( hd==NULL || hd->vtype != FITS_VINT )
	return(FIZIP_ERR_CONFLICTING_HEADERS);
		/* discrepancy: NAXIS1 is not present 		     */

 cmtot=hd->vint;
 crtot=fits_tapeblock_size(cmtot);

 fizip_restore_headers(header,hbuff);

 if ( fits_table_get_params(header,&ft) )
	return(FIZIP_ERR_CONFLICTING_HEADERS);
 ctot=ft.nrow*ft.rowsize;
 atot=ctot+1;
 rtot=fits_tapeblock_size(ctot);
 
 imgrd=(char *)malloc(rtot);
 imgcc=(char *)malloc(crtot);

 cb_read(rdparam,imgcc,crtot);
 w=ci->decompress(imgcc,cmtot,imgrd,&atot);

 free(imgcc);
 if ( w || atot != ctot )
  {	free(imgrd);
	return(FIZIP_ERR_DECOMPRESS);
  }

 if ( ctot<rtot )	memset(imgrd+ctot,0,rtot-ctot);

 fits_headerset_write_cb(cb_write,wrparam,header);
 cb_write(wrparam,imgrd,rtot);

 free(imgrd);
	
 return(0);
}

/** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** **/

int fizip_compress_bintable
	(int (*cb_read )(void *,void *,int),void *rdparam,
	 int (*cb_write)(void *,void *,int),void *wrparam,
	 fitsheaderset *header,fitsbtable *fb,compressinfo *ci,
	 void *param,compresstune *ct)
{
 char		hbuff[80];
 int		btot,ctot,cmps,wtot,nread;
 fitsheader	*hd;
 char		*imgrd,*imgcc;

 if ( header==NULL || fb==NULL || ci==NULL || cb_read==NULL || cb_write==NULL )
	return(FIZIP_ERR_INTERNAL);

 fizip_save_bintable_headers(fb,hbuff);
 fits_headerset_set_string(header,fiziphdrs[HDR_FIZIPHDR],FITS_SH_FIRST,hbuff,NULL);
 btot=fb->nrow*fb->rowsize;

 ctot=fits_tapeblock_size(btot);

 imgrd=(char *)malloc(ctot);
 nread=cb_read(rdparam,imgrd,ctot);
 if ( nread < ctot )
	return(FIZIP_ERR_UNEXPECTED_EOF);
 cmps=ctot;
 imgcc=(char *)malloc(cmps);

 if ( ci->compress(imgrd,btot,imgcc,&cmps,param) )
	return(FIZIP_ERR_COMPRESS);
 
 hd=fits_headerset_get_header(header,"NAXIS1",0);
 hd->vint=1;
 hd=fits_headerset_get_header(header,"NAXIS2",0);
 hd->vint=cmps;
 hd=fits_headerset_get_header(header,"TFIELDS",0);
 hd->vint=1;
 hd=fits_headerset_get_header(header,"TFORM1",0);
 strcpy(hd->vstr,"1B");

 fits_headerset_write_cb(cb_write,wrparam,header);
 
 wtot=fits_tapeblock_size(cmps);
 imgcc=(char *)realloc(imgcc,wtot);
 if ( wtot>cmps )	memset(imgcc+cmps,0,wtot-cmps);

 cb_write(wrparam,imgcc,wtot);

 return(0);
}

int fizip_decompress_bintable
	(int (*cb_read )(void *,void *,int),void *rdparam,
	 int (*cb_write)(void *,void *,int),void *wrparam,
	 fitsheaderset *header,char *hbuff,compressinfo *ci)
{
 fitsheader	*hd;
 char		*imgcc,*imgrd;
 int		cmtot,crtot,ctot,atot,rtot,w;
 fitsbtable	fb;

 hd=fits_headerset_get_uniq_header(header,"NAXIS");
 if ( hd==NULL || hd->vtype != FITS_VINT || hd->vint != 2 )
 	return(FIZIP_ERR_CONFLICTING_HEADERS);
		/* discrepancy: NAXIS is not present or it is not 2  */
 hd=fits_headerset_get_uniq_header(header,"NAXIS1");
 if ( hd==NULL || hd->vtype != FITS_VINT || hd->vint != 1 )
	return(FIZIP_ERR_CONFLICTING_HEADERS);
		/* discrepancy: NAXIS1 is not present or it is not 1 */
 hd=fits_headerset_get_uniq_header(header,"NAXIS2");
 if ( hd==NULL || hd->vtype != FITS_VINT )
	return(FIZIP_ERR_CONFLICTING_HEADERS);
		/* discrepancy: NAXIS1 is not present 		     */

 cmtot=hd->vint;
 crtot=fits_tapeblock_size(cmtot);

 fizip_restore_headers(header,hbuff);

 if ( fits_bintable_get_params(header,&fb) )
	return(FIZIP_ERR_CONFLICTING_HEADERS);

 ctot=fb.nrow*fb.rowsize;
 atot=ctot+1;
 rtot=fits_tapeblock_size(ctot);
 
 imgrd=(char *)malloc(rtot);
 imgcc=(char *)malloc(crtot);

 cb_read(rdparam,imgcc,crtot);
 w=ci->decompress(imgcc,cmtot,imgrd,&atot);

 free(imgcc);
 if ( w || atot != ctot )
  {	free(imgrd);
	return(FIZIP_ERR_DECOMPRESS);
  }

 if ( ctot<rtot )	memset(imgrd+ctot,0,rtot-ctot);

 fits_headerset_write_cb(cb_write,wrparam,header);
 cb_write(wrparam,imgrd,rtot);

 free(imgrd);
	
 return(0);
}

/*****************************************************************************/

int fizip_compress_extensions(
	int (*cb_read )(void *,void *,int),void *rdprm,
	int (*cb_write)(void *,void *,int),void *wrprm,
	compressinfo *ci,void *param,compresstune *ct)
{
 fitsextension	xtn;
 int		ret;
 
 while ( 1 )
  {	fits_headerset_reset(&xtn.header);
	fits_headerset_read_cb(cb_read,rdprm,&xtn.header);
	if ( xtn.header.nhdr<=0 ) break;	/* End-of-stream occured */
	xtn.type=fits_headerset_is_extension(&xtn.header);
	switch ( xtn.type )
	 {   case FITS_EXT_IMAGE:
		if ( fits_image_get_params(&xtn.header,&xtn.x.i) )
			ret=FIZIP_ERR_CONFLICTING_HEADERS;
		else
			ret=fizip_compress_image(cb_read,rdprm,cb_write,wrprm,&xtn.header,&xtn.x.i,ci,param,ct);
		break;
	     case FITS_EXT_TABLE:
		if ( fits_table_get_params(&xtn.header,&xtn.x.t) )
			ret=FIZIP_ERR_CONFLICTING_HEADERS;
		else
			ret=fizip_compress_table(cb_read,rdprm,cb_write,wrprm,&xtn.header,&xtn.x.t,ci,param,ct);
		break;
	     case FITS_EXT_BINTABLE:
		if ( fits_bintable_get_params(&xtn.header,&xtn.x.b) )
			ret=FIZIP_ERR_CONFLICTING_HEADERS;
		else
			ret=fizip_compress_bintable(cb_read,rdprm,cb_write,wrprm,&xtn.header,&xtn.x.b,ci,param,ct);
		break;
	     default:
		ret=FIZIP_ERR_UNSUPPORTED_EXTENSION;
		break;
	 }
	fits_headerset_free(&xtn.header);
	if ( ret )	return(ret);
  }

 return(0);
}

int fizip_compress(
	int (*cb_read )(void *,void *,int),void *rdprm,
	int (*cb_write)(void *,void *,int),void *wrprm,
	int method,void *param,compresstune *ct)
{
 fits		*img;
 int		cnt,i,is_primary,ret;
 compressinfo	*ci; 
 fitsheader	*hx;

 ci=fizip_get_compressinfo(method);
 if ( ci==NULL )	return(FIZIP_ERR_UNIMPLEMENTED_METHOD);

 img=fits_create();
 fits_headerset_reset(&img->header);

 fits_headerset_read_cb(cb_read,rdprm,&img->header); 
 if ( img->header.nhdr<=0 )
	return(FIZIP_ERR_UNEXPECTED_EOF);
 for ( i=0 ; fiziphdrs[i] != NULL ; i++ )
  {	cnt=fits_headerset_get_count(&img->header,fiziphdrs[HDR_FIZIP]);
	if ( cnt>0 )	/* already compressed */
		return(FIZIP_ERR_ALREADY_COMPRESSED); 
  }

 fits_headerset_set_string(&img->header,fiziphdrs[HDR_FIZIPVRS],FITS_SH_BEGIN,ci->version(),"Compressor version");
 fits_headerset_set_string(&img->header,fiziphdrs[HDR_FIZIPMTD],FITS_SH_BEGIN,ci->name,"Compression method");
 fits_headerset_set_boolean(&img->header,fiziphdrs[HDR_FIZIP],FITS_SH_BEGIN,1,"Compressed by fizip");

 is_primary=!fits_image_get_params(&img->header,&img->i);
 if ( is_primary )
  {	ret=fizip_compress_image(cb_read,rdprm,cb_write,wrprm,&img->header,&img->i,ci,param,ct);
	if ( ret )
	 {	fits_free(img);return(ret); 		}
  }
 else
  {	fits_headerset_write_cb(cb_write,wrprm,&img->header);				}

 hx=fits_headerset_get_uniq_header(&img->header,"EXTEND");
 if ( hx != NULL && hx->vtype==FITS_VBOOLEAN && hx->vint )
  {	ret=fizip_compress_extensions(cb_read,rdprm,cb_write,wrprm,ci,param,ct);
	if ( ret )
	 {	fits_free(img);return(ret);		}
  }
 
 fits_free(img);

 return(0);
}

int fizip_decompress_extensions(
	int (*cb_read )(void *,void *,int),void *rdprm,
	int (*cb_write)(void *,void *,int),void *wrprm,
	compressinfo *ci)
{
 fitsextension	xtn;
 fitsheader	*hd;
 char		hbuff[80];
 int		ret;
 
 while ( 1 )
  {	fits_headerset_reset(&xtn.header);
	fits_headerset_read_cb(cb_read,rdprm,&xtn.header);
	if ( xtn.header.nhdr<=0 ) break;	/* End-of-stream occured */
	xtn.type=fits_headerset_is_extension(&xtn.header);
	hd=fits_headerset_get_uniq_header(&xtn.header,fiziphdrs[HDR_FIZIPHDR]);
	if ( hd != NULL && hd->vtype==FITS_VSTR )
	 {	strcpy(hbuff,hd->vstr);			}
	switch ( xtn.type )
	 {   case FITS_EXT_IMAGE:
		fits_headerset_delete_all(&xtn.header,fiziphdrs[HDR_FIZIPHDR]);
		ret=fizip_decompress_image(cb_read,rdprm,cb_write,wrprm,&xtn.header,hbuff,ci);
		break;
	     case FITS_EXT_TABLE:
		fits_headerset_delete_all(&xtn.header,fiziphdrs[HDR_FIZIPHDR]);
		ret=fizip_decompress_table(cb_read,rdprm,cb_write,wrprm,&xtn.header,hbuff,ci);
		break;
	     case FITS_EXT_BINTABLE:
		fits_headerset_delete_all(&xtn.header,fiziphdrs[HDR_FIZIPHDR]);
		ret=fizip_decompress_bintable(cb_read,rdprm,cb_write,wrprm,&xtn.header,hbuff,ci);
		break;
	     default:
		ret=FIZIP_ERR_UNSUPPORTED_EXTENSION;
		break;
	 }
	fits_headerset_free(&xtn.header);
	if ( ret )	return(ret);
  }

 return(0);
}


int fizip_decompress(int (*cb_read )(void *,void *,int),void *rdprm,
		     int (*cb_write)(void *,void *,int),void *wrprm)
{
 fits		*img;
 compressinfo	*ci;
 fitsheader	*hd,*hx;
 char		hbuff[80];
 int		ret;

 img=fits_create();
 fits_headerset_reset(&img->header);
 fits_headerset_read_cb(cb_read,rdprm,&img->header); 

 hd=fits_headerset_get_uniq_header(&img->header,fiziphdrs[HDR_FIZIP]);
 if ( hd==NULL || hd->vtype != FITS_VBOOLEAN || (!hd->vint) )
  {	fits_free(img);		/* not compressed by fizip */
	return(FIZIP_ERR_NOT_COMPRESSED);
  }	
 hd=fits_headerset_get_uniq_header(&img->header,fiziphdrs[HDR_FIZIPMTD]);
 if ( hd==NULL || hd->vtype != FITS_VSTR )
  {	fits_free(img);		/* ... discrepancy 1	   */
	return(FIZIP_ERR_CONFLICTING_HEADERS);
  }	
 ci=fizip_get_compressinfo_byname(hd->vstr);
 if ( ci==NULL )		/* unknown compr. method   */
  {	fits_free(img);
	return(FIZIP_ERR_UNIMPLEMENTED_METHOD);
  }	
 hd=fits_headerset_get_uniq_header(&img->header,fiziphdrs[HDR_FIZIPVRS]);
 if ( hd==NULL || hd->vtype != FITS_VSTR )
  {	fits_free(img);		/* ... discrepancy 3	   */
	return(FIZIP_ERR_CONFLICTING_HEADERS);
  }	

 fits_headerset_delete_all(&img->header,fiziphdrs[HDR_FIZIP]);
 fits_headerset_delete_all(&img->header,fiziphdrs[HDR_FIZIPMTD]);
 fits_headerset_delete_all(&img->header,fiziphdrs[HDR_FIZIPVRS]);

 hd=fits_headerset_get_uniq_header(&img->header,fiziphdrs[HDR_FIZIPHDR]);
 if ( hd != NULL && hd->vtype==FITS_VSTR )	/* pri.image is compressed */
  {	strcpy(hbuff,hd->vstr);
	fits_headerset_delete_all(&img->header,fiziphdrs[HDR_FIZIPHDR]);
	ret=fizip_decompress_image(cb_read,rdprm,cb_write,wrprm,&img->header,hbuff,ci);
  }
 else	/* no primary image */
  {	fits_headerset_delete_all(&img->header,fiziphdrs[HDR_FIZIPHDR]);
	fits_headerset_write_cb(cb_write,wrprm,&img->header);
	ret=0;
  }

 if ( ret )
  {	fits_free(img);return(ret);		}

 hx=fits_headerset_get_uniq_header(&img->header,"EXTEND");
 if ( hx != NULL && hx->vtype==FITS_VBOOLEAN && hx->vint )
  {     ret=fizip_decompress_extensions(cb_read,rdprm,cb_write,wrprm,ci);
	if ( ret )
	 {	fits_free(img);return(ret);		}
  }

 fits_free(img);
 
 return(0);
}

/*****************************************************************************/
                               
